# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file RobotMoveLocusRDB.py
 @brief ロボット移動軌跡表示で使用するRDB
 @date $Date$

 @author East Japan Accounting Center Co.,Ltd.

 (c)2017 The University of Aizu
 This software is released under the New BSD License.

"""
import sys
import time
sys.path.append(".")

# Import RTM module
import RTC
import OpenRTM_aist


# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>
import os
import pymongo
import numpy as np
import csv
import json
import datetime as dt
import dateutil.parser

""" message definition """
messege_text_01 = ("onInitialize")
messege_text_02 = ("onFinalize")
messege_text_03 = ("onActivated")
messege_text_04 = ("onDeactivated")
messege_text_05 = ("Heart beat by RDB")
messege_text_06 = (
    "Error was detected. \
Please reset the RTC by the RT System Editor.")
messege_text_07 = ("Input database search condition")
messege_text_10 = ("Date data is wrong. Please confirm.")
messege_text_11 = ("Search conditions are wrong. Please confirm.")
messege_text_12 = ("!write fail - file ")
messege_text_13 = ("!write okay - file ")
messege_text_14 = (
    "There is no data \
that meet the search conditions.")
messege_text_15 = ("Connected to the MongoDB")
messege_text_16 = ("Not connected to the MongoDB")
messege_text_17 = ("Could not connect to the MongoDB.")
messege_text_18 = ("The specified number of lists exceeded.")
messege_text_19 = ("onAborting")
messege_text_20 = ("RobotID is not on the list. Please confirm.")
messege_text_21 = ("directory make failed. \
    export csv file current directory.")
messege_text_22 = ("export directry path :")
messege_text_23 = ("export directory make")
messege_text_24 = ("datetime validity check OK")
messege_text_25 = ("datetime string check OK")

""" cloud MongoDB find Key"""
key_dp_robotID_string = "dp_robotID.string"
key_server_time = "server_time.string"
key_dp_droneGPS = "dp_droneGPS"
key_dp_gyro = "dp_gyro"

""" search condition """
cmd_start_tim = "StartDateTime"
cmd_end_tim = "EndDateTime"
cmd_db_col = "DBCollection"
cmd_robo_id = "RobotID"

""" mongoDB """
MONGO_DB_NAME = ("dbrtc")
MONGO_DB_HOST_NAME = "localhost"
MONGO_DB_PORT = 27017

""" export csv file header """
hedder_server_time = "server_time"
csvheader_list = (hedder_server_time, key_dp_droneGPS, key_dp_gyro)
""" find robot ID list """
robo_id_list = ("drone1", "drone2", "drone3", "drone4",
                "drone5", "drone6", "drone7", "drone8",
                "gorone1", "gorone2", "gorone3", "gorone4",
                "gorone5", "gorone6", "gorone7", "gorone8",
                "enryu1", "spider1")
""" search condition list """
cmd_param_list = (cmd_start_tim, cmd_end_tim, cmd_db_col, cmd_robo_id)
""" RTC Status """
RTC_STS_INIT = "RDB_STS_INIT"
RTC_STS_NO_ACTION = "RDB_STS_NO_ACTION"
RTC_STS_READ = "RDB_STS_READ"
""" onExecute Stage """
EXE_STAGE_INIT = "execute_stage_init"
EXE_STAGE_NO_0 = "execute_stage0"
EXE_STAGE_NO_1 = "execute_stage1"
EXE_STAGE_NO_2 = "execute_stage2"


# This module's spesification
# <rtc-template block="module_spec">
robotmovelocusrdb_spec = ["implementation_id", "RobotMoveLocusRDB",
                          "type_name",         "RobotMoveLocusRDB",
                          "description",       "ロボット移動軌跡表示で使用するRDB",
                          "version",           "1.0.0",
                          "vendor",            "EAC",
                          "category",          "Category",
                          "activity_type",     "STATIC",
                          "max_instance",      "1",
                          "language",          "Python",
                          "lang_type",         "SCRIPT",
                          "conf.default.export_path", "SensorDataExport/",
                          "conf.default.robot_GPS_key", "dp_droneGPS",
                          "conf.default.robot_gyro_key", "dp_gyro",
                          "conf.__widget__.export_path", "text",
                          "conf.__widget__.robot_GPS_key", "text",
                          "conf.__widget__.robot_gyro_key", "text",
                          ""]
# </rtc-template>

#
# @class RobotMoveLocusRDB
# @brief ロボット移動軌跡表示で使用するRDB
#
# 別紙"機能仕様書_ロボット移動軌跡RDB RTC"参照
#
# 別紙"機能仕様書_ロボット移動軌跡RDB RTC"参照
#
# 別紙"機能仕様書_ロボット移動軌跡RDB RTC"参照
#
#


class RobotMoveLocusRDB(OpenRTM_aist.DataFlowComponentBase):

    #
    # @brief constructor
    # @param manager Maneger Object
    #

    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_inDBSearchParam = RTC.TimedString(RTC.Time(0, 0), "")
        """
        パラメータ（collection、ロボットID、取得開始日時、終了日時）の設定を行う
        """
        self._dp_DBSearchParamIn = OpenRTM_aist.InPort(
            "dp_DBSearchParam", self._d_inDBSearchParam)
        self._d_outRobotMoveLocusRDBSts = RTC.TimedString(RTC.Time(0, 0), "")
        """
        ロボット移動軌跡RDBの状態を通知する
        """
        self._dp_RobotMoveLocusRDBStsOut = OpenRTM_aist.OutPort(
            "dp_RobotMoveLocusRDBSts", self._d_outRobotMoveLocusRDBSts)

        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">
        """

        - Name:  export_path
        - DefaultValue: SensorDataExport/
        """
        self._export_path = ['SensorDataExport/']
        """

        - Name:  robot_GPS_key
        - DefaultValue: dp_droneGPS
        """
        self._robot_GPS_key = ['dp_droneGPS']

        """

        - Name:  robot_gyro_key
        - DefaultValue: dp_gyro
        """
        self._robot_gyro_key = ['dp_gyro']

        # </rtc-template>

    #
    # ロボット移動軌跡表示RDBのInitialize処理
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry()
    #
    # @return RTC::ReturnCode_t
    #
    #
    def onInitialize(self):
        # Bind variables and configuration variable
        self.bindParameter(
            "export_path", self._export_path, "SensorDataExport/")
        self.bindParameter("robot_GPS_key", self._robot_GPS_key, "dp_droneGPS")
        self.bindParameter("robot_gyro_key", self._robot_gyro_key, "dp_gyro")

        # Set InPort buffers
        self.addInPort("dp_DBSearchParam", self._dp_DBSearchParamIn)

        # Set OutPort buffers
        self.addOutPort(
            "dp_RobotMoveLocusRDBSts", self._dp_RobotMoveLocusRDBStsOut)

        # Set service provider to Ports

        # Set service consumers to Ports

        # Set CORBA Service Ports
        self.onInitializeFunc()
        self._MongoDBConnection()
        print messege_text_01
        return RTC.RTC_OK

        #
        # ロボット移動軌跡表示RDBのFinalize処理
        #
        # The finalize action (on ALIVE->END transition)
        # formaer rtc_exiting_entry()
        #
        # @return RTC::ReturnCode_t

        #
    def onFinalize(self):
        print messege_text_02
        return RTC.RTC_OK

    #
    #
    # The startup action when ExecutionContext startup
    # former rtc_starting_entry()
    #
    # @param ec_id target ExecutionContext Id
    #
    # @return RTC::ReturnCode_t
    #
    #
    # def onStartup(self, ec_id):
    #
    # return RTC.RTC_OK

    #
    #
    # The shutdown action when ExecutionContext stop
    # former rtc_stopping_entry()
    #
    # @param ec_id target ExecutionContext Id
    #
    # @return RTC::ReturnCode_t
    #
    #
    # def onShutdown(self, ec_id):
    #
    # return RTC.RTC_OK

        #
        # ロボット移動軌跡表示RDBのActivate処理
        #
        # The activated action (Active state entry action)
        # former rtc_active_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onActivated(self, ec_id):
        print messege_text_03
        self.onActivatedFunc()

        return RTC.RTC_OK

        #
        # ロボット移動軌跡表示RDBのDeactivate処理
        #
        # The deactivated action (Active state exit action)
        # former rtc_active_exit()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onDeactivated(self, ec_id):
        self.onDeactivatedFunc()
        print messege_text_04
        return RTC.RTC_OK

        #
        # ロボット移動軌跡表示RDBのExecute処理
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onExecute(self, ec_id):
        self.onExectuteFunc()
        return RTC.RTC_OK

        #
        # ロボット移動軌跡表示RDBのAborting処理
        #
        # The aborting action when main logic error occurred.
        # former rtc_aborting_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onAborting(self, ec_id):
        print messege_text_19
        self.onAbortingFunc()
        return RTC.RTC_OK

    #
    #
    # The error action in ERROR state
    # former rtc_error_do()
    #
    # @param ec_id target ExecutionContext Id
    #
    # @return RTC::ReturnCode_t
    #
    #
    # def onError(self, ec_id):
    #
    # return RTC.RTC_OK

        #
        # ロボット移動軌跡表示RDBのReset処理
        #
        # The reset action that is invoked resetting
        # This is same but different the former rtc_init_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onReset(self, ec_id):

        return RTC.RTC_OK

    #
    #
    # The state update action that is invoked after onExecute() action
    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #
    # @param ec_id target ExecutionContext Id
    #
    # @return RTC::ReturnCode_t
    #

    #
    # def onStateUpdate(self, ec_id):
    #
    # return RTC.RTC_OK

    #
    #
    # The action that is invoked when execution context's rate is changed
    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #
    # @param ec_id target ExecutionContext Id
    #
    # @return RTC::ReturnCode_t
    #
    #
    # def onRateChanged(self, ec_id):
    #
    # return RTC.RTC_OK
    # =========================================================================
    # Function Name   :onInitializeFunc
    # Description     :This function is an "onInitialize process"
    #                  run-time initialization process.
    # Argument        :nothing special
    # Return          :nothing special
    # =========================================================================
    def onInitializeFunc(self):
        self.serchcond = {}
        self.db_connection_sts = False
        self.rows = []
        self.rtc_sts = RTC_STS_INIT
        self.onAbortingDetect = 0
        self.nextstage = EXE_STAGE_INIT
        self.filetimestamp = ""
        self.dir_path = ""

    # =========================================================================
    # Function Name   :onActivatedFunc
    # Description     :This function is an "onActivated process"
    #                  run-time initialization process.
    # Argument        :onAborting process detection inf,
    # Return          :nothing special
    # =========================================================================
    def onActivatedFunc(self):
        self.rtc_sts_write(RTC_STS_INIT)
        self.reset_next_stage()
        self.make_export_dir()
        if self.onAbortingDetect == 1:
            self.onInitializeFunc()
            self._MongoDBConnection()
        else:
            self.onAbortingDetect = 0

    # =========================================================================
    # Function Name   :onExecuteFunc
    # Description     :This function is the main function that is executed
    #                  from "onExecute" function.
    # Argument        :input-port of the RTC
    # Return          :nothing special
    # =========================================================================
    def onExectuteFunc(self):
        if self.nextstage == EXE_STAGE_NO_0:
            self.execute_stage0()
        elif self.nextstage == EXE_STAGE_NO_1:
            self.execute_stage1()
        elif self.nextstage == EXE_STAGE_NO_2:
            self.execute_stage2()
        else:
            pass

    # =========================================================================
    # Function Name   :execute_stage0
    # Description     :This function is onExecute function stage0
    # Argument        :nothing special
    # Return          :nothing special
    # =========================================================================
    def execute_stage0(self):
        self.rtc_sts_write(RTC_STS_NO_ACTION)
        self.set_next_stage()

    # =========================================================================
    # Function Name   :execute_stage1
    # Description     :This function is onExecute function stage1
    # Argument        :nothing special
    # Return          :nothing special
    # =========================================================================
    def execute_stage1(self):
        if self._dp_DBSearchParamIn.isNew():
            print messege_text_07
            s_param = self.readSerchParamPort()
            if not s_param:
                print messege_text_11
                return
            if not self.serch_parameter_chk(s_param):
                print messege_text_11
                return
            if self.mongo_read():
                self.set_next_stage()
            else:
                self.reset_next_stage()

    # =========================================================================
    # Function Name   :execute_stage2
    # Description     :This function is onExecute function stage2
    # Argument        :nothing special
    # Return          :nothing special
    # =========================================================================
    def execute_stage2(self):
        self.export_csv()
        self.reset_next_stage()

    # =========================================================================
    # Function Name   :set_next_stage
    # Description     :This function is set stage number
    # Argument        :nothing special
    # Return          :nothing special
    # =========================================================================
    def set_next_stage(self):
        if self.nextstage == EXE_STAGE_NO_0:
            self.nextstage = EXE_STAGE_NO_1
        elif self.nextstage == EXE_STAGE_NO_1:
            self.nextstage = EXE_STAGE_NO_2
        else:
            self.nextstage = EXE_STAGE_INIT

    # =========================================================================
    # Function Name   :reset_next_stage
    # Description     :This function is
    # Argument        :nothing special
    # Return          :nothing special
    # =========================================================================
    def reset_next_stage(self):
        self.nextstage = EXE_STAGE_NO_0

    # =========================================================================
    # Function Name   :onDeactivatedFunc
    # Description     :This function is an "onDeactivated process"
    #                  run-time initialization process.
    # Argument        :nothing special
    # Return          :nothing special
    # =========================================================================
    def onDeactivatedFunc(self):
        print messege_text_04

    # =========================================================================
    # Function Name   :onAbortingFunc
    # Description     :This function is an "onAborting process"
    #                  run-time initialization process.
    # Argument        :nothing special
    # Return          :nothing special
    # =========================================================================
    def onAbortingFunc(self):
        self.onAbortingDetect = 1

    # =========================================================================
    # Function Name   :readSerchParamPort
    # Description     :This function is read RTC data Port
    # Argument        :nothing special
    # Return          :serch parameter (type dict) or False
    # =========================================================================
    def readSerchParamPort(self):
        json_str = self._dp_DBSearchParamIn.read().data
        try:
            json_dict = json.loads(json_str)
        except Exception:
            return False
        return (json_dict)

    # =========================================================================
    # Function Name   :rtc_sts_write
    # Description     :This function is write RTC data Port
    # Argument        :RTC status
    # Return          :nothing special
    # =========================================================================
    def rtc_sts_write(self, sts):
        self._d_outRobotMoveLocusRDBSts.data = sts
        ret = self._dp_RobotMoveLocusRDBStsOut.write()
        if ret:
            print "RDB RTC STS"
            print self._d_outRobotMoveLocusRDBSts

    # =========================================================================
    # Function Name   :_MongoDBConnection
    # Description     :This function execute the connection
    #                  of MongoDB and the RTC.
    # Argument        :nothing special
    # Return          :nothing special
    # =========================================================================
    def _MongoDBConnection(self):
        self.db_connection_sts = False
        try:
            self.client = pymongo.MongoClient(
                MONGO_DB_HOST_NAME, MONGO_DB_PORT)
        except Exception as err:
            print messege_text_17
            print err.message
        else:
            self.db_connection_sts = True
            self.db = self.client[MONGO_DB_NAME]
            print messege_text_15
            print ("db name is = " + self.db.name)

    # =========================================================================
    # Function Name   :mongo_read
    # Description     :
    # Argument        :nothing special
    # Return          :information of normal end(True) or abnormal end(False).
    # =========================================================================
    def mongo_read(self):
        return_value = False
        if not self.get_db_connection_sts():
            print messege_text_16
            return return_value
        self.rtc_sts_write(RTC_STS_READ)
        if self.find_mongoDB():
            return_value = True
        else:
            print messege_text_14
        return return_value

    # =========================================================================
    # Function Name   :get_db_connection_sts
    # Description     :This function returns the database connection status.
    # Argument        :nothing special
    # Return          :information of normal end(True) or abnormal end(False).
    # =========================================================================
    def get_db_connection_sts(self):
        return_value = False

        if self.db_connection_sts:
            return_value = True

        return return_value

    # =========================================================================
    # Function Name   :serch_parameter_chk
    # Description     :This function is check to serch parameter.
    # Argument        :serch_parameter
    # Return          :information of normal end(True) or abnormal end(False).
    # =========================================================================
    def serch_parameter_chk(self, s_param):
        return_value = False
        keyslist = set(s_param.keys())
        if keyslist != set(cmd_param_list):
            print messege_text_11
            return return_value
        if s_param[cmd_robo_id] not in robo_id_list:
            print messege_text_20
            return return_value
        if self.check_string_of_datetime(s_param):
            self.command_set(s_param)
            return_value = True
        return return_value

    # =========================================================================
    # Function Name   :command_set
    # Description     :This function is replicate serch parameter.
    # Argument        :serch parameter.
    # Return          :nothing special
    # =========================================================================
    def command_set(self, s_param):
        self.serchcond.clear()
        self.serchcond = dict(s_param)

    # =========================================================================
    # Function Name   :check_string_of_datetime
    # Description     :This function confirm the string of dates and time.
    # Argument        :date, time
    # Return          :information of normal end(True) or abnormal end(False).
    # =========================================================================
    def check_string_of_datetime(self, cmd):
        return_value = False

        try:
            s_date_parse = dateutil.parser.parse(cmd[cmd_start_tim])
            d1 = s_date_parse.strftime("%Y%m%d-%H%M%S")
        except ValueError:
            print messege_text_10
        except Exception as err:
            print err.message
        else:
            try:
                e_date_parse = dateutil.parser.parse(cmd[cmd_end_tim])
                d2 = e_date_parse.strftime("%H%M%S")
            except ValueError:
                print messege_text_10
            except Exception as err:
                print err.message
            else:
                print messege_text_25
                self.filetimestamp = (d1 + "-" + d2)
                return_value = self.validity_check_of_datetime(
                    s_date_parse, e_date_parse)

        return return_value

    # =========================================================================
    # Function Name   :validity_check_of_datetime
    # Description     :This function to check
    #                  whether the valid information of the date and time.
    # Argument        :start datetime, end datetime
    # Return          :information of normal end(True) or abnormal end(False).
    # =========================================================================
    def validity_check_of_datetime(self, s_date_parse, e_date_parse):
        return_value = False

        if e_date_parse > s_date_parse:
            return_value = True
            print messege_text_24
        else:
            print messege_text_10
        return return_value

    # =========================================================================
    # Function Name   :find_mongoDB
    # Description     :This function is find mongoDB
    # Argument        :nothing special
    # Return          :information of normal end(True) or abnormal end(False).
    # =========================================================================
    def find_mongoDB(self):
        key_list = list((hedder_server_time,
                        self._robot_GPS_key[0],
                        self._robot_gyro_key[0]))
        co = self.db[self.serchcond[cmd_db_col]]

        del self.rows[:]
        self.rows.append(list(csvheader_list))
        s_start = self.serchcond[cmd_start_tim]
        s_end = self.serchcond[cmd_end_tim]
        s_robot_id = self.serchcond[cmd_robo_id]
        query = {key_dp_robotID_string: s_robot_id,
                 key_server_time: {"$gte": s_start, "$lte": s_end}}

        f_data = None
        for f_data in co.find(query).sort(key_server_time):
            row = []
            for i in key_list:
                row.append(f_data.get(i, np.nan))
            self.rows.append(row)
        if f_data is None:
            return False

        return True

    # =========================================================================
    # Function Name   :make_export_dir
    # Description     :This function is make directory to export csv file.
    # Argument        :nothing special
    # Return          :nothing special
    # =========================================================================
    def make_export_dir(self):
        d_date = dt.date.today().strftime('%Y%m%d')
        d_path = (self._export_path[0] + d_date + "/")
        self.dir_path = ""
        try:
            ret = os.path.isdir(d_path)
        except Exception as err:
            print messege_text_21
        else:
            if not ret:
                try:
                    os.makedirs(d_path)
                    self.dir_path = d_path
                    print messege_text_23
                    print messege_text_22 + self.dir_path
                except Exception as err:
                    print messege_text_21
                    print err
                else:
                    pass
            else:
                self.dir_path = d_path
                print messege_text_22 + self.dir_path

    # =========================================================================
    # Function Name   :export_csv
    # Description     :This function is export csv file.
    # Argument        :nothing special
    # Return          :information of normal end(True) or abnormal end(False).
    # =========================================================================
    def export_csv(self):
        retvalue = False
        filename = self.genarate_fileName()
        print "write data [[header],[data]...]"
        print self.rows

        fullpath = (self.dir_path + filename)

        try:
            with open(fullpath, 'w') as wf:
                writer = csv.writer(wf, lineterminator='\n')
                writer.writerows(self.rows)
        except Exception as err:
            print messege_text_12
            print err.message
        else:
            print messege_text_13
            print "Export File name: " + filename
            retvalue = True
        del self.rows[:]
        return retvalue

    # =========================================================================
    # Function Name   :genarate_fileName
    # Description     :
    # Argument        :nothing special
    # Return          :filename
    # =========================================================================
    def genarate_fileName(self):
        s_db_col = self.serchcond[cmd_db_col]
        s_robot_id = self.serchcond[cmd_robo_id]

        filename = (s_db_col + "_" + s_robot_id +
                    "_" + self.filetimestamp + ".csv")

        return filename


def RobotMoveLocusRDBInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=robotmovelocusrdb_spec)
    manager.registerFactory(profile,
                            RobotMoveLocusRDB,
                            OpenRTM_aist.Delete)


def MyModuleInit(manager):
    RobotMoveLocusRDBInit(manager)

    # Create a component
    comp = manager.createComponent("RobotMoveLocusRDB")


def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager()


if __name__ == "__main__":
    main()
